from typing import List, Union

import numpy as np
import pandas as pd
from scipy.stats import qmc

from estival.priors import PriorDict
from estival.sampling import tools as esamptools

from dataclasses import dataclass


@dataclass
class PriorSizeInfo:
    sizes: List[int]
    tot_size: int
    offsets: List[slice]


class SampleTypes:
    INPUT = "input"
    DICT = "dict"
    LIST_OF_DICTS = "list_of_dicts"
    ARRAY = "array"
    PANDAS = "pandas"
    SAMPLEITERATOR = "sample"


def constrain(sample, priors: PriorDict, bounds=0.99):
    # return {k:np.clip(sample[k], *priors[k].bounds(bounds)) for k,v in sample.items()}
    def constrain_op(sample, prior):
        return np.clip(sample, *prior.bounds(bounds))

    return _process_samples_for_priors(sample, priors, constrain_op)


def ppf(sample, priors: PriorDict):
    def ppf_op(sample, prior):
        return prior.ppf(sample)

    return _process_samples_for_priors(sample, priors, ppf_op)


def cdf(sample, priors: PriorDict):
    def cdf_op(sample, prior):
        return prior.cdf(sample)

    return _process_samples_for_priors(sample, priors, cdf_op)


def get_prior_sizeinfo(priors) -> PriorSizeInfo:
    sizes = [p.size for k, p in priors.items()]
    tot_size = sum(sizes)
    offsets = [0] + list(np.cumsum(sizes))
    offset_idx = []
    for i in range(len(offsets) - 1):
        if sizes[i] == 1:
            offset_idx.append(offsets[i])
        else:
            offset_idx.append(slice(offsets[i], offsets[i + 1]))
    # offset_idx = [slice(offsets[i], offsets[i + 1]) for i in range(len(offsets) - 1)]
    return PriorSizeInfo(sizes, tot_size, offset_idx)


def _process_samples_for_priors(sample, priors: PriorDict, op_func):
    size_info = get_prior_sizeinfo(priors)
    psize = size_info.tot_size
    poffsets = size_info.offsets
    if isinstance(sample, np.ndarray):
        shape = sample.shape
        if len(shape) == 1:
            if len(sample) == psize:
                return np.array(
                    [op_func(sample[poffsets[i]], p) for i, p in enumerate(priors.values())]
                )
            else:
                raise ValueError("Input sample must be same size as priors")
        elif len(shape) == 2:
            if shape[1] == psize:
                out_arr = np.empty_like(sample)
                priors_list = list(priors.values())
                for i, pidx in enumerate(poffsets):
                    # for i, psamp_set in enumerate(sample.T):
                    psamp_set = sample[:, pidx]
                    out_arr[:, pidx] = op_func(psamp_set, priors_list[i])
                return out_arr
            else:
                raise ValueError(
                    "Shape mismatch: Could not broadcast input sample to priors", shape
                )
        else:
            raise ValueError(f"Invalid shape {shape} for sample")
    elif isinstance(sample, dict):
        return {k: op_func(v, priors[k]) for k, v in sample.items()}
    elif isinstance(sample, pd.Series):
        return pd.Series({k: op_func(v, priors[k]) for k, v in sample.items()})
    elif isinstance(sample, pd.DataFrame):
        out_df = pd.DataFrame(index=sample.index)
        for c in sample.columns:
            out_df[c] = op_func(sample[c].to_numpy(), priors[c])
        return out_df
    elif isinstance(sample, list):
        assert all([isinstance(subsample, dict) for subsample in sample])
        return [_process_samples_for_priors(subsample, priors, op_func) for subsample in sample]


def convert_sample_type(sample, priors, target_type: str):
    if target_type == SampleTypes.INPUT:
        return sample

    size_info = get_prior_sizeinfo(priors)
    psize = size_info.tot_size

    if target_type == SampleTypes.SAMPLEITERATOR:
        if isinstance(sample, np.ndarray):
            return esamptools.SampleIterator.from_array(sample, priors)
        else:
            return esamptools.validate_samplecontainer(sample)

    if isinstance(sample, np.ndarray):
        if target_type == SampleTypes.ARRAY:
            return sample
        elif target_type == SampleTypes.LIST_OF_DICTS:
            if len(sample.shape) == 1:
                sample = sample.reshape((len(sample), 1))
            if len(sample.shape) == 2:
                return [
                    {k: subsample[size_info.offsets[i]] for i, k in enumerate(priors)}
                    for subsample in sample
                ]
            else:
                raise ValueError(
                    "Shape mismatch: Could not broadcast input sample to priors", sample.shape
                )
        elif target_type == SampleTypes.DICT:
            assert len(sample.shape) == 1
            assert len(sample) == psize
            return {k: sample[i] for i, k in enumerate(priors)}
        elif target_type == SampleTypes.PANDAS:
            df = pd.DataFrame(sample, columns=priors)
            df.index.name = "sample"
            return df
        else:
            raise ValueError(f"Target type {target_type} not supported for array inputs")
    elif isinstance(sample, list):
        assert isinstance(sample[0], dict)
        if target_type == SampleTypes.ARRAY:
            return _lod_to_arr(sample, priors)
        elif target_type == SampleTypes.PANDAS:
            df = pd.DataFrame(_lod_to_arr(sample, priors), columns=priors)
            df.index.name = "sample"
            return df
    elif isinstance(sample, pd.DataFrame):
        if target_type == SampleTypes.LIST_OF_DICTS:
            return [v.to_dict() for _, v in sample.iterrows()]
        elif target_type == SampleTypes.ARRAY:
            return sample.to_numpy()
        elif target_type == SampleTypes.PANDAS:
            return sample
    else:
        sample = esamptools.validate_samplecontainer(sample)
        if target_type == SampleTypes.LIST_OF_DICTS:
            return [v for _, v in sample.iterrows()]  # type: ignore
        elif target_type == SampleTypes.PANDAS:
            return pd.DataFrame(sample.components, index=sample.index)

    raise TypeError(
        "Unsupported combination of input type and target type", type(sample), target_type
    )


def _lod_to_arr(in_lod, priors):
    assert len(in_lod[0]) == len(priors)
    out_arr = np.empty((len(in_lod), len(in_lod[0])))
    for i, in_dict in enumerate(in_lod):
        for j, k in enumerate(priors):
            out_arr[i, j] = in_dict[k]
    return out_arr


def _lod_to_si(lod):
    ref_dict = lod[0]
    components = {}
    for k in ref_dict:
        components[k] = np.stack([samp[k] for samp in lod])
    return esamptools.SampleIterator(components)


class SampledPriorsManager:
    def __init__(self, priors):
        self.priors = priors
        self.size_info = get_prior_sizeinfo(priors)

    def constrain(self, sample, bounds=0.99, ret_type=SampleTypes.INPUT):
        return convert_sample_type(constrain(sample, self.priors, bounds), self.priors, ret_type)

    def ppf(self, sample, ret_type=SampleTypes.INPUT):
        return convert_sample_type(ppf(sample, self.priors), self.priors, ret_type)

    def cdf(self, sample, ret_type=SampleTypes.INPUT):
        return convert_sample_type(cdf(sample, self.priors), self.priors, ret_type)

    def convert(self, sample, ret_type):
        return convert_sample_type(sample, self.priors, ret_type)

    def lhs(self, n_samples: int, out_type="sample", clamp=0.99):
        lhs = qmc.LatinHypercube(self.size_info.tot_size)
        samples = lhs.random(n_samples)
        clamp_offset = (1.0 - clamp) * 0.5
        samples = np.clip(samples, clamp_offset, 1.0 - clamp_offset)
        resampled = self.ppf(samples)
        return self.convert(resampled, out_type)

    def sobol(self, n_samples: int, out_type="sample", clamp=0.99):
        sobol = qmc.Sobol(self.size_info.tot_size)
        samples = sobol.random(n_samples)
        clamp_offset = (1.0 - clamp) * 0.5
        samples = np.clip(samples, clamp_offset, 1.0 - clamp_offset)
        resampled = self.ppf(samples)
        return self.convert(resampled, out_type)

    def uniform(self, n_samples: int, out_type="sample", clamp=0.99):
        samples = np.random.uniform(size=(n_samples, self.size_info.tot_size))
        clamp_offset = (1.0 - clamp) * 0.5
        samples = np.clip(samples, clamp_offset, 1.0 - clamp_offset)
        resampled = self.ppf(samples)
        return self.convert(resampled, out_type)



from inspect import istraceback
from typing import Tuple, Dict, Optional, Union
from multiprocessing import cpu_count
from dataclasses import dataclass

import pandas as pd
from arviz import InferenceData
import numpy as np
from pandas.core.frame import dataclasses_to_dicts

import xarray

from estival.model import BayesianCompartmentalModel, ResultsData
from estival.utils.parallel import map_parallel
from estival.utils.sample import convert_sample_type, get_prior_sizeinfo, _lod_to_si

SampleIndex = Tuple[int, int]
ParamDict = Dict[str, float]
SampleContainer = Union[pd.DataFrame, "SampleIterator", xarray.Dataset]


@dataclass
class SampledResults:
    results: pd.DataFrame
    extras: Optional[pd.DataFrame]


@dataclass
class _PriorStub:
    size: int


def likelihood_extras_for_idata(
    idata: InferenceData,
    bcm: BayesianCompartmentalModel,
    num_workers: Optional[int] = None,
    exec_mode: str = "thread",
) -> pd.DataFrame:
    """Calculate the likelihood extras (ll,lprior,lpost + per-target) for all
    samples in supplied InferenceData, returning a DataFrame.

    Note - input InferenceData must be the full (unburnt) idata

    Args:
        idata: The InferenceData to sample
        bcm: The BayesianCompartmentalModel (must be the same BCM used to generate idata)
        num_workers: Number of multiprocessing workers to use; defaults to cpu_count/2

    Returns:
        A DataFrame with index (chain, draw) and columns being the keys in ResultsData.extras
            - Use df.reset_index(level="chain").pivot(columns="chain") to move chain into column multiindex
    """
    num_workers = num_workers or int(cpu_count() / 2)

    accepted_s = idata["sample_stats"].accepted.copy()
    # Handle pathological cases where we've burnt out the first accepted sample
    accepted_s[:, 0] = True

    accepted_df = accepted_s.to_dataframe()

    accepted_index = accepted_df[accepted_df["accepted"] == True].index

    accept_mask = accepted_s.data
    posterior_t = idata["posterior"].transpose("chain", "draw", ...)

    components = {}
    for dv in posterior_t.data_vars:
        components[dv] = posterior_t[dv].data[accept_mask]

    accepted_si = SampleIterator(components, index=accepted_index)
    # Get the likelihood extras for all accepted samples - this spins up a multiprocessing pool
    # pres = sample_likelihood_extras_mp(bcm, accepted_samples_df, n_workers)

    extras_df = likelihood_extras_for_samples(accepted_si, bcm, num_workers, exec_mode=exec_mode)

    # Collate this into an array - it's much much faster than dealing with pandas directly
    tmp_extras = np.empty((len(accepted_df), extras_df.shape[-1]))

    # This value should never get used - we know something went wrong if the accepted field if it did
    last_good_sample_idx = "IndexNotSet"

    for i, (idx, accepted_s) in enumerate(accepted_df.iterrows()):
        # Extract the bool from the Series
        accepted = accepted_s["accepted"]
        # Update the index if this sample is accepted - otherwise we'll
        # store the previous known good sample (ala MCMC)
        if accepted:
            last_good_sample_idx = idx
        tmp_extras[i] = extras_df.loc[last_good_sample_idx]

    # Create a DataFrame with the full index of the idata
    # This has a lot of redundant information, but it's still only a few Mb and
    # makes lookup _so_ much easier...
    filled_edf = pd.DataFrame(
        index=accepted_df.index, columns=extras_df.columns, data=tmp_extras, dtype=float
    )

    return filled_edf


def _extras_df_from_pres(pres, is_full_data=False, index_names=("chain", "draw")) -> pd.DataFrame:
    extras_dict = {
        "logposterior": {},
        "logprior": {},
        "loglikelihood": {},
    }

    base_fields = list(extras_dict)

    for idx, res in pres:
        if is_full_data:
            extras = res.extras
        else:
            extras = res
        for field in base_fields:
            extras_dict[field][idx] = float(extras[field])
        for k, v in extras["ll_components"].items():
            extras_dict.setdefault("ll_" + k, {})
            extras_dict["ll_" + k][idx] = float(v)

    extras_df = pd.DataFrame(extras_dict)
    extras_df.index = extras_df.index.set_names(index_names)  # pyright: ignore

    return extras_df


def likelihood_extras_for_samples(
    samples: SampleContainer,
    bcm: BayesianCompartmentalModel,
    num_workers: Optional[int] = None,
    exec_mode: Optional[str] = "thread",
) -> pd.DataFrame:
    def get_sample_extras(sample_params: Tuple[SampleIndex, ParamDict]) -> Tuple[SampleIndex, dict]:
        """Run the BCM for a given set of parameters, and return its extras dictionary
        (likelihood, posterior etc)

        Args:
            sample_params: The parameter set to sample (indexed by chain,draw)

        Returns:
            A tuple of SampleIndex and the ResultsData.extras dictionary
        """

        idx, params = sample_params
        res = bcm.run(params, include_extras=True)
        return idx, res.extras

    samples = validate_samplecontainer(samples)

    pres = map_parallel(get_sample_extras, samples.iterrows(), num_workers, mode=exec_mode)

    if isinstance(samples.index, pd.MultiIndex):
        levels = samples.index.names
        unstack_levels = list(range(len(levels)))
    else:
        unstack_levels = (0,)
        if hasattr(samples.index, "name"):
            name = samples.index.name or "sample"  # type: ignore
        else:
            name = "sample"
        levels = (name,)

    return _extras_df_from_pres(pres, False, index_names=levels)


def model_results_for_samples(
    samples: SampleContainer,
    bcm: BayesianCompartmentalModel,
    include_extras: bool = False,
    num_workers: Optional[int] = None,
    exec_mode: Optional[str] = "thread",
) -> SampledResults:
    def get_model_results(
        sample_params: Tuple[SampleIndex, ParamDict]
    ) -> Tuple[SampleIndex, ResultsData]:
        """Run the BCM for a given set of parameters, and return its extras dictionary
        (likelihood, posterior etc)

        Args:
            sample_params: The parameter set to sample (indexed by chain,draw)

        Returns:
            A tuple of SampleIndex and the ResultsData.extras dictionary
        """

        idx, params = sample_params
        res = bcm.run(params, include_extras=include_extras)
        return idx, res

    samples = validate_samplecontainer(samples)

    pres = map_parallel(get_model_results, samples.iterrows(), num_workers, mode=exec_mode)

    df = pd.concat([p[1].derived_outputs for p in pres], keys=[p[0] for p in pres])

    if isinstance(samples.index, pd.MultiIndex):
        levels = samples.index.names
        unstack_levels = list(range(len(levels)))
    else:
        unstack_levels = (0,)
        if hasattr(samples.index, "name"):
            name = samples.index.name or "sample"  # type: ignore
        else:
            name = "sample"
        levels = (name,)

    df: pd.DataFrame = df.sort_index().unstack(level=unstack_levels)  # type: ignore
    df.columns.set_names(["variable", *levels], inplace=True)
    df.index.set_names("time", inplace=True)

    if include_extras:
        extras_df: pd.DataFrame = _extras_df_from_pres(
            pres, True, index_names=levels
        ).sort_index()  # type:ignore
        return SampledResults(df, extras_df)
    else:
        return SampledResults(df, None)


def quantiles_for_results(results_df: pd.DataFrame, quantiles: Tuple[float]) -> pd.DataFrame:
    """Summary

    Args:
        results_df: DataFrame with layout equivalent to model_results_for_samples output
        quantiles: Quantiles to compute [0.0,1.0]

    Returns:
        pd.DataFrame: DataFrame with time as index and [variable, quantile] as columns
    """
    columns = pd.MultiIndex.from_product(
        (results_df.columns.levels[0], quantiles), names=["variable", "quantile"]  # type: ignore
    )
    udf = pd.DataFrame(index=results_df.index, columns=columns)

    for variable in results_df.columns.levels[0]:  # type: ignore
        udf[variable] = results_df[variable].quantile(quantiles, axis=1).T  # type: ignore

    return udf


class IndexGetter:
    def __init__(self, func):
        self.func = func

    def __getitem__(self, key):
        return self.func(key)


class SampleIterator:
    """A simple container storing dicts of arrays, providing a means to iterate over
    the array items (and returning a dict of the items at each index)
    Designed to be a drop-in replacement for pd.DataFrame.iterrows (but supporting multidimensional arrays)
    """

    def __init__(self, components: dict, index=None):
        self.components = components
        self.clen = self._calc_component_length()
        if index is None:
            self.index = pd.RangeIndex(self.clen, name="sample")
        else:
            assert (
                idxlen := len(index)
            ) == self.clen, f"Index length {idxlen} not equal to component length {self.clen}"
            self.index = index
        self._idxidx = pd.Series(index=self.index, data=range(len(self.index)))

        self.iloc = IndexGetter(self._get_by_iloc)

        self._priors_stub = self._build_priorsize_table()

    def _calc_component_length(self) -> int:
        clen = -1
        for k, cval in self.components.items():
            if clen == -1:
                clen = len(cval)
            else:
                assert len(cval) == clen, f"Length mismatch for {k} ({len(cval)}), should be {clen}"
        return clen

    def _build_priorsize_table(self) -> dict:
        """
        A dictionary of stub objects resembling BasePrior that contain only a size field
        """

        priors_stub = {}
        for k, v in self.components.items():
            refv = v[0]
            if isinstance(refv, np.ndarray):
                size = len(refv)
            else:
                size = 1
            priors_stub[k] = _PriorStub(size)
        return priors_stub

    def __iter__(self):
        for i in range(self.clen):
            out = {}
            for k, v in self.components.items():
                out[k] = v[i]
            yield out

    def iterrows(self):
        for i in range(self.clen):
            out = {}
            for k, v in self.components.items():
                out[k] = v[i]
            yield self.index[i], out

    def __getitem__(self, idx):
        out = {}
        if isinstance(idx, pd.Index):
            data_idx = self._idxidx[idx]
            for k, v in self.components.items():
                out[k] = v[data_idx]
            return SampleIterator(out, index=idx)
        else:
            for k, v in self.components.items():
                out[k] = v[idx]
            return out

    def _get_by_iloc(self, key):
        return self[self.index[key]]

    def convert(self, target_type: str):
        return convert_sample_type(self, self._priors_stub, target_type=target_type)

    @classmethod
    def from_array(cls, in_array, priors):
        size_info = get_prior_sizeinfo(priors)
        components = {k: in_array[:, size_info.offsets[i]] for i, k in enumerate(priors)}
        return cls(components)


def xarray_to_sampleiterator(in_data: xarray.Dataset):
    if list(in_data.dims)[0] == "sample":
        index = in_data.sample.to_index()
        data_t = in_data.transpose("sample", ...)
        n_idxdim = 1
    elif list(in_data.dims)[:2] == ["chain", "draw"]:
        index = pd.MultiIndex.from_product(
            [in_data.coords["chain"].to_index(), in_data.coords["draw"].to_index()]
        )
        data_t = in_data.transpose("chain", "draw", ...)
        n_idxdim = 2
    else:
        raise KeyError("Incompatible dimensions ")

    components = {}
    for dv in in_data.data_vars:
        dvar = data_t[dv]

        sample_shape = int(np.prod(dvar.data.shape[:n_idxdim]))
        var_shape = dvar.data.shape[n_idxdim:]
        final_shape = [sample_shape] + list(var_shape)

        components[dv] = dvar.data.reshape(final_shape)

    si = SampleIterator(components, index=index)
    return si


def idata_to_sampleiterator(in_data: InferenceData, group="posterior"):
    return xarray_to_sampleiterator(in_data[group])


def dataframe_to_sampleiterator(in_data: pd.DataFrame):
    components = {c: in_data[c].to_numpy() for c in in_data.columns}  # type: ignore
    return SampleIterator(components, index=in_data.index)


def validate_samplecontainer(in_data: SampleContainer) -> Union[SampleIterator, pd.DataFrame]:
    if isinstance(in_data, SampledResults):
        return in_data.results
    if isinstance(in_data, InferenceData):
        return idata_to_sampleiterator(in_data)
    elif isinstance(in_data, xarray.Dataset):
        return xarray_to_sampleiterator(in_data)
    elif isinstance(in_data, SampleIterator):
        return in_data
    elif isinstance(in_data, pd.DataFrame):
        return dataframe_to_sampleiterator(in_data)
    elif isinstance(in_data, list):
        return _lod_to_si(in_data)
    else:
        raise TypeError("Unsupported type", in_data)



from typing import Callable, Iterable, Optional, Type

import multiprocessing as mp

from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

import cloudpickle


# initialize a worker in the process pool
def process_init_cloudpickle(custom_cpkl):
    """Initialize a mp.Process with custom cloudpickled data

    Args:
        custom_cpkl: The cloudpickle.dumps output of the custom data
    """
    # declare global variable
    global custom_data
    # assign the global variable
    custom_data = cloudpickle.loads(custom_cpkl)


def generic_cpkl_worker(*args):
    """Worker function to be used with multiprocessing pools, where
    custom_data is a global initialized to a function, usually via
    process_init_cloudpickle

    Returns:
        Return type of the custom global function
    """
    global custom_data
    run_func = custom_data

    return run_func(*args)


def map_parallel(
    run_func: Callable,
    input_iterator: Iterable,
    n_workers: Optional[int] = None,
    mode: Optional[str] = "process",
):
    """Map the values of input_iterator over a function run_func, using n_workers parallel workers
    Defaults to ProcessPoolExecutor; for non-Python-bound tasks, 'thread'

    Args:
        run_func: The function to call over the mapped inputs
        input_iterator: An iterable containing the values to map
        n_workers: Number of processes used by Pool
        mode: ProcessExecutor type; either 'thread' or 'process'.  'process' is required for
              non-thread-safe tasks, while 'thread' is usually faster for small jax-heavy tasks

    Returns:
        A list of values return by run_func
    """

    if n_workers is None:
        n_workers = int(mp.cpu_count())

    if mode is None:
        mode = "thread"

    if mode == "process":
        with ProcessPoolExecutor(  # type: ignore
            n_workers, initializer=process_init_cloudpickle, initargs=(cloudpickle.dumps(run_func),)
        ) as pool:
            pres = pool.map(generic_cpkl_worker, input_iterator)
            pres = [p for p in pres]
    elif mode == "thread":
        with ThreadPoolExecutor(n_workers) as pool:  # type: ignore
            pres = pool.map(run_func, input_iterator)
            pres = [p for p in pres]
    else:
        raise ValueError("Mode must be one of ['thread', 'process']")

    return pres
